## This code is created by Jakob Hansen-Tangen and Sindre Sangnes as work to obtain a master thesis 
## within Managerial economics and Operations Research at the Department of Industrial Economics and Technology 
## at the Norwegian University of Science and Technology (NTNU)

## Date 7/6-13

## The code is the MINLP formulation of the production system at Petrobras' production FPSO, P35
## The model is testet with real data from the asset

## The code includes an extra constraint to prevent some symmetry. 
## This constraint prevents pipeline 'prod2' from being routed to separator 'sep2', and is addet at the end of the model.
## This constraint is originally not part of the general MINLP model of the asset.

########################################################################################
# MINLP formulation
########################################################################################

#Defining problem sets
set MANIFOLDS; 									#set of all manifolds, manifold m
set MANIFOLDSTOP within MANIFOLDS; 				#set of topside manifolds, manifold m
set MANIFOLDSSUB within MANIFOLDS; 				#set of subsea manifolds, manifold m
set WELLS; 										#set of all wells, well j
set WELLSM{MANIFOLDS} within WELLS;				#set of wells, well j,connected to manifold m
set PIPELINES; 									#set of all pipelines, line l
set PIPELINESM{MANIFOLDS} within PIPELINES;		#set of pipelines, line l, connected to manifold m
set PHASES; 									#set of phases, phase o, g, and w
set SEPARATORS; 								#set of separators, separator s
set ALPHAS;										#set of alphas from polynomial oil flow curve fitting, alpha a
set BETAS;										#set of betas from polynomial pressure drop curve fitting, beta b

#Defining problem parameters
param sep_p{SEPARATORS}; 										#separator pressure, the same for all separators
param fpso_GLcap;												#gaslift capacity for platform
param GL_comp_p;												#pressure at gaslift compressor
param Max_flow_well{m in MANIFOLDS, j in WELLS};				#maximum oil flow out well j connected to manifold m
param Max_LG_injected{m in MANIFOLDS, j in WELLS};				#upper limit on lift gas injected into each well
param Min_LG_injected{m in MANIFOLDS,j in WELLS};				#lower limit on lift gas injected into each well
param Max_p_well{m in MANIFOLDS, j in WELLS};					#upper limit on wellhead pressure in each well
param Min_p_well{m in MANIFOLDS,j in WELLS};					#lower limit for wellhead pressure in each well
param alpha{m in MANIFOLDS, j in WELLS, a in ALPHAS};			#function describing flow from well j connected to topside manifold m
param beta{m in MANIFOLDSSUB, l in PIPELINES, b in BETAS};		#function describing pressure drop in pipeline l,connected to subsea manifold m
param GOR{m in MANIFOLDS, j in WELLS};							#gas to oil ratio for well j connected to manifold m
param WC{m in MANIFOLDS, j in WELLS};							#short term water cut for well j connected to manifold m
param Gas_Capacity;												#total gas handling capacity for platform (all separators)
param Sep_Liquid_Capacity{SEPARATORS};							#separator capacity for liquid flow for separator s
param Sep_Water_Capacity{SEPARATORS};							#separator capacity for water flow for separator s

#Defining continous variables
var flow_well{m in MANIFOLDS, j in WELLSM[m], p in PHASES} >= 0;			#flow of phase p from well j to manifold m
var flow_pipeline{m in MANIFOLDSSUB, l in PIPELINESM[m], p in PHASES} >= 0;	#flow of phase p in pipeline l connected to manifold m
var flow_LG{m in MANIFOLDS, j in WELLSM[m]} >= 0;							#flow of liftgas injected in well j connected to manifold m
var p_well{m in MANIFOLDS, j in WELLSM[m]} >= 0;							#pressure at well head for well j connected to manifold m
var p_pipeline{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;					#pressure before separator inlet in pipeline l,connected to manifold m
var p_manifold{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;					#pressure in pipeline l at manifold m
var p_drop{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;						#pressure drop in pipeline l,connected to manifold m

#Defining binary variables
var man_connect{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]} binary;				#1 if well j is routed to pipeline l,connected to manifold m,0 otherwise
var sub_sep{m in MANIFOLDSSUB, l in PIPELINESM[m], s in SEPARATORS} binary;					#1 if pipeline l connected to subsea manifold m is routed to separator s,0 otherwise
var top_sep{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS} binary;						#1 if well j connected to topside manifold m is routed to separator s,0 otherwise
var LG_bin{m in MANIFOLDS, j in WELLSM[m]} binary;											#1 if lift gas is injected to well j, connected to manifold m, 0 otherwise

#Objective function
maximize Oil_Prod:																#Maximizing total oil production
	sum{s in SEPARATORS, m in MANIFOLDSTOP, j in WELLSM[m]}(flow_well[m,j,'o']*top_sep[m,j,s])
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]} flow_pipeline[m,l,'o'];

	
subject to MaxWellFlow{m in MANIFOLDS, j in WELLSM[m]}:										# flow out of a well cannot exceed a given maximum flow out of well
	flow_well[m,j,'o'] <= Max_flow_well[m,j];

subject to LiftGasInterval1{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]}:			# amount of lift gas injected into a well is restricted by a lower limit if injected
	flow_LG[m,j] >=  Min_LG_injected[m,j]*man_connect[m,j,l]; 								

subject to LiftGasInterval2{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS}:			# amount of lift gas injected into a well is restricted by a lower limit if injected
	flow_LG[m,j] >=  Min_LG_injected[m,j]*top_sep[m,j,s]; 	

subject to LiftGasInterval3{m in MANIFOLDS, j in WELLSM[m]}:								# amount of lift gas injected into a well is restricted by an upper limit if injected
	flow_LG[m,j] <= Max_LG_injected[m,j];								
	
subject to WHPressureInterval{m in MANIFOLDS, j in WELLSM[m]}:								# restricting well head pressure to be below its upper limit
	Min_p_well[m,j] <= p_well[m,j] <= Max_p_well[m,j];	

subject to OilFlow{m in MANIFOLDS,j in WELLSM[m]}:											# function describing oil flow out of wells
	flow_well[m,j,'o'] = alpha[m,j,'a1'] + alpha[m,j,'a2']*p_well[m,j] 
	+ alpha[m,j,'a3']*flow_LG[m,j] + alpha[m,j,'a4']*p_well[m,j]*flow_LG[m,j]
	+ alpha[m,j,'a5']*p_well[m,j]*p_well[m,j] + alpha[m,j,'a6']*flow_LG[m,j]*flow_LG[m,j];

subject to GasFlow{m in MANIFOLDS, j in WELLSM[m]}:											# relationship between oil and gas flow out of wells
	flow_well[m,j,'g'] = flow_well[m,j,'o']*GOR[m,j] + flow_LG[m,j];

subject to WaterFlow{m in MANIFOLDS, j in WELLSM[m]}:										# relationship between oil and water flow out of wells
	flow_well[m,j,'w'] = (flow_well[m,j,'o']*WC[m,j])/(1-WC[m,j]);

subject to pCon1{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]}:						# pressure at manifold is less than well pressure for opened wells
	p_manifold[m,l]*man_connect[m,j,l] <= p_well[m,j];
	
subject to pDropCon1{m in MANIFOLDSSUB, l in PIPELINESM[m]}:									# pressure drop equals the difference between pressure in pipeline and in manifold
	p_drop[m,l] = p_manifold[m,l] - p_pipeline[m,l];

subject to pDropCon2{m in MANIFOLDSSUB, l in PIPELINESM[m]}:									# function describing pressure drop in pipelines
	p_drop[m,l] = beta[m,l,'b1']*log(flow_pipeline[m,l,'o']+1) 
	+ beta[m,l,'b2']*log(flow_pipeline[m,l,'g']+1) + beta[m,l,'b3']*log(flow_pipeline[m,l,'w']+1)
	+ beta[m,l,'b4']*(log(flow_pipeline[m,l,'o']+1))^2 
	+ beta[m,l,'b5']*(log(flow_pipeline[m,l,'g']+1))^2 
	+ beta[m,l,'b6']*(log(flow_pipeline[m,l,'w']+1))^2 
	+ beta[m,l,'b7']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'g']+1) 
	+ beta[m,l,'b8']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'w']+1)
	+ beta[m,l,'b9']*log(flow_pipeline[m,l,'g']+1)*log(flow_pipeline[m,l,'w']+1) 
	+ beta[m,l,'b10']*(log(flow_pipeline[m,l,'o']+1))^3 
	+ beta[m,l,'b11']*(log(flow_pipeline[m,l,'g']+1))^3 
	+ beta[m,l,'b12']*(log(flow_pipeline[m,l,'w']+1))^3 
	+ beta[m,l,'b13']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'g']+1)*log(flow_pipeline[m,l,'w']+1);

subject to ManBinCon{m in MANIFOLDSSUB, j in WELLSM[m]}:									# a well can only be connected to one pipeline
	sum{l in PIPELINESM[m]}man_connect[m,j,l] <= 1;
	
subject to FlowCon1{m in MANIFOLDSSUB, l in PIPELINESM[m], p in PHASES}:					# flow in a pipeline equals the flow of all open wells connected to that line
	flow_pipeline[m,l,p] = sum{j in WELLSM[m]}flow_well[m,j,p]*man_connect[m,j,l];
	
subject to pCon2{s in SEPARATORS, m in MANIFOLDSSUB, l in PIPELINESM[m]}:					# pressure in separator is less than pressure in pipeline for subsea manifolds
	sep_p[s]*sub_sep[m,l,s] <= p_pipeline[m,l];
	
subject to pCon3{s in SEPARATORS, m in MANIFOLDSTOP, j in WELLSM[m]}:						# pressure in separator is less than pressure in well for topside manifolds
	sep_p[s]*top_sep[m,j,s] <= p_well[m,j];

subject to sub_BinCon{m in MANIFOLDSSUB, l in PIPELINESM[m]}:								# a pipeline can only be connected to one separator
	sum{s in SEPARATORS}sub_sep[m,l,s] = 1;

subject to top_BinCon{m in MANIFOLDSTOP, j in WELLSM[m]}:									# a well, connected to the topside manifold can only be connected to one separator
	sum{s in SEPARATORS}top_sep[m,j,s] <= 1;
	
subject to LGCap:																			# the sum of gaslift injection in all wells cannot exceed liftgas capacity of the platform
	sum{m in MANIFOLDSSUB, j in WELLSM[m]}(flow_LG[m,j])
	+ sum{m in MANIFOLDSTOP, j in WELLSM[m]}(flow_LG[m,j]) <= fpso_GLcap;

subject to TotalGasCapacity:																#total gas flow through system must be less than then handling capacity
	sum{s in SEPARATORS, m in MANIFOLDSTOP, j in WELLSM[m]}(flow_well[m,j,'g']*top_sep[m,j,s])
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}(flow_pipeline[m,l,'g']) <= Gas_Capacity;
	
subject to SeparatorLiquidCapacity{s in SEPARATORS}:										#liquid flow must be less than the liquid handling capacity at each separator
	sum{m in MANIFOLDSTOP, j in WELLSM[m]}(flow_well[m,j,'o']*top_sep[m,j,s])
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}(flow_pipeline[m,l,'o']*sub_sep[m,l,s])
	+ sum{m in MANIFOLDSTOP, j in WELLSM[m]}(flow_well[m,j,'w']*top_sep[m,j,s])
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}(flow_pipeline[m,l,'w']*sub_sep[m,l,s]) <= Sep_Liquid_Capacity[s];
	
subject to SeparatorWaterCapacity{s in SEPARATORS}:											#each separator has a water separation capacity
	sum{m in MANIFOLDSTOP, j in WELLSM[m]}(flow_well[m,j,'w']*top_sep[m,j,s])
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}(flow_pipeline[m,l,'w']*sub_sep[m,l,s]) <= Sep_Water_Capacity[s];

subject to SymmetryCon1:												# additional constraint added to reduce symmetry of the problem
	sub_sep['sub1', 'prod1', 'sep2'] = 0;