## NLP-formulation created by Sindre Sangnes and Jakob Hansen_Tangen
## Used for all nodes in the logic branch and bound algorithm

#Defining problem sets
set MANIFOLDS; 									#set of all manifolds, manifold m
set MANIFOLDSTOP within MANIFOLDS; 				#set of topside manifolds, manifold m
set MANIFOLDSSUB within MANIFOLDS; 				#set of subsea manifolds, manifold m
set WELLS; 										#set of all wells, well j
set WELLSM{MANIFOLDS} within WELLS;				#set of wells, well j,connected to manifold m
set PIPELINES; 									#set of all pipelines, line l
set PIPELINESM{MANIFOLDS} within PIPELINES;		#set of pipelines, line l, connected to manifold m
set PHASES; 									#set of phases, phase o, g, and w
set SEPARATORS; 								#set of separators, separator s
set ALPHAS;										#set of alphas from polynomial oil flow curve fitting, alpha a
set BETAS;										#set of betas from polynomial pressure drop curve fitting, beta b

#Defining problem parameters
param sep_p{SEPARATORS}; 										#separator pressure, the same for all separators
param sep_cap{p in PHASES, s in SEPARATORS}; 					#handling capacity for phase p in separator s
param fpso_GLcap;												#gaslift capacity for platform
param GL_comp_p;												#pressure at gaslift compressor
param Max_flow_well{m in MANIFOLDS, j in WELLS};				#maximum oil flow out well j connected to manifold m
param Max_LG_injected{m in MANIFOLDS, j in WELLS};				#upper bound on lift gas injected into each well
param Max_p_well{m in MANIFOLDS, j in WELLS};					#upper bound on wellhead pressure in each well
param alpha{m in MANIFOLDS, j in WELLS, a in ALPHAS};			#function describing flow from well j connected to topside manifold m
param beta{m in MANIFOLDSSUB, l in PIPELINES, b in BETAS};		#function describing pressure drop in pipeline l,connected to subsea manifold m
param GOR{m in MANIFOLDS, j in WELLS};							#gas to oil ratio for well j connected to manifold m
param WC{m in MANIFOLDS, j in WELLS};							#short term water cut for well j connected to manifold m
param Gas_Capacity;												#total gas handling capacity for platform (all separators)
param Sep_Water_Capacity{SEPARATORS};							#water handling capacity for separator s
param Sep_Liquid_Capacity{SEPARATORS};							#separator capacity for liquids for separator s

param Min_LG_injected{m in MANIFOLDS, j in WELLS};
param Min_p_well{m in MANIFOLDS, j in WELLS};

#Defining continous variables
var flow_well{m in MANIFOLDS, j in WELLSM[m], p in PHASES} >= 0;			#flow of phase p from well j to manifold m
var flow_pipeline{m in MANIFOLDSSUB, l in PIPELINESM[m], p in PHASES} >= 0;	#flow of phase p in pipeline l connected to (subsea) manifold m
var flow_LG{m in MANIFOLDS, j in WELLSM[m]} >= 0;							#flow of liftgas injected in well j connected to manifold m
var p_well{m in MANIFOLDS, j in WELLSM[m]} >= 0;							#pressure at well head for well j connected to manifold m
var p_pipeline{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;					#pressure before separator inlet in pipeline l,connected to manifold m
var p_manifold{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;					#pressure in pipeline l at manifold m
var p_drop{m in MANIFOLDSSUB, l in PIPELINESM[m]} >= 0;						#pressure drop in pipeline l,connected to manifold m

#Extra variables
var flow_manifold{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINES, p in PHASES} >= 0;		#flow of phase p from well j to pipeline l connected through subsea manifold m
var subflow_sep{m in MANIFOLDSSUB, l in PIPELINESM[m], s in SEPARATORS, p in PHASES} >= 0;	#flow of phase p into separator s from pipeline l connected to subsea manifold m
var topflow_sep{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS, p in PHASES} >= 0;		#flow of phase p into separator s from well j connected to manifold m
var flow_LG_manifold{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]} >= 0;			#flow of liftgas from well j,connected to subsea manifold m,to pipeline l
var topflow_LG{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS} >= 0;					#flow of litgas from well j, connected to topside manifold m, to separator s

#Objective function
maximize Oil_Prod:
	sum{m in MANIFOLDSTOP, j in WELLSM[m]} flow_well[m,j,'o'] 							#Maximizing total oil production
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]} flow_pipeline[m,l,'o'];


subject to MaximumWellFlow{m in MANIFOLDS, j in WELLSM[m]}:
	flow_well[m,j,'o'] <= Max_flow_well[m,j];											# flow out of a well cannot exceed a given maximum flow out of well
	
subject to MaxLiftGasInjected{m in MANIFOLDS, j in WELLSM[m]}:
	flow_LG[m,j] <= Max_LG_injected[m,j];												# lift gas injected into each well might be restricted by an upper bound
	
subject to MaxWellHeadPressure{m in MANIFOLDS, j in WELLSM[m]}:
	Min_p_well[m,j] <= p_well[m,j] <= Max_p_well[m,j];														# restricting well head pressure to be below its upper bound

subject to OilFlow{m in MANIFOLDS,j in WELLSM[m]}:										# function describing oil flow out of wells
	flow_well[m,j,'o'] = alpha[m,j,'a1'] + alpha[m,j,'a2']*p_well[m,j] 
	+ alpha[m,j,'a3']*flow_LG[m,j] + alpha[m,j,'a4']*p_well[m,j]*flow_LG[m,j]
	+ alpha[m,j,'a5']*p_well[m,j]*p_well[m,j] + alpha[m,j,'a6']*flow_LG[m,j]*flow_LG[m,j];


subject to pDropCon1{m in MANIFOLDSSUB, l in PIPELINESM[m]}:							# pressure drop equals the difference between pressure in pipeline and in manifold
	p_drop[m,l]	= p_manifold[m,l] - p_pipeline[m,l];

subject to pDropCon2{m in MANIFOLDSSUB, l in PIPELINESM[m]}:							# function describing pressure drop in pipelines
	p_drop[m,l] = beta[m,l,'b1']*log(flow_pipeline[m,l,'o']+1) 
	+ beta[m,l,'b2']*log(flow_pipeline[m,l,'g']+1) + beta[m,l,'b3']*log(flow_pipeline[m,l,'w']+1)
	+ beta[m,l,'b4']*(log(flow_pipeline[m,l,'o']+1))^2 
	+ beta[m,l,'b5']*(log(flow_pipeline[m,l,'g']+1))^2 
	+ beta[m,l,'b6']*(log(flow_pipeline[m,l,'w']+1))^2 
	+ beta[m,l,'b7']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'g']+1) 
	+ beta[m,l,'b8']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'w']+1)
	+ beta[m,l,'b9']*log(flow_pipeline[m,l,'g']+1)*log(flow_pipeline[m,l,'w']+1) 
	+ beta[m,l,'b10']*(log(flow_pipeline[m,l,'o']+1))^3 
	+ beta[m,l,'b11']*(log(flow_pipeline[m,l,'g']+1))^3 
	+ beta[m,l,'b12']*(log(flow_pipeline[m,l,'w']+1))^3 
	+ beta[m,l,'b13']*log(flow_pipeline[m,l,'o']+1)*log(flow_pipeline[m,l,'g']+1)*log(flow_pipeline[m,l,'w']+1);
	
subject to LGCap:																		# the sum of gaslift injection in all wells cannot exceed liftgas capacity of the platform
	sum{m in MANIFOLDS, j in WELLSM[m]}flow_LG[m,j] <= fpso_GLcap;

subject to balanceCon1{m in MANIFOLDSSUB, l in PIPELINESM[m], p in PHASES}:				# flow in pipeline l equals the sum of flow from all wells into that pipeline
	flow_pipeline[m,l,p] = sum{j in WELLSM[m]}flow_manifold[m,j,l,p];
	
subject to balanceCon2{m in MANIFOLDSSUB, j in WELLSM[m]}:								# flow out of well j equals the sum of flows in all pipelines from that well
	flow_well[m,j,'o'] = sum{l in PIPELINESM[m]}flow_manifold[m,j,l,'o'];
	
subject to balanceCon3{m in MANIFOLDSSUB, l in PIPELINESM[m], p in PHASES}:				# flow in pipeline l equals the sum of flow to all separators from that pipeline
	flow_pipeline[m,l,p] = sum{s in SEPARATORS}subflow_sep[m,l,s,p];
	
#subject to balanceCon4{m in MANIFOLDSTOP, j in WELLSM[m], p in PHASES}:				# flow out of well j equals the sum of flows to all separators from that well
#	flow_well[m,j,p] = sum{s in SEPARATORS}topflow_sep[m,j,s,p];

subject to LiftGasBalance1{m in MANIFOLDSSUB, j in WELLSM[m]}:							# flow of liftgas to each well j connected to subseaman m equals the flow routed to pipelines
	flow_LG[m,j] = sum{l in PIPELINESM[m]}(flow_LG_manifold[m,j,l]);
	
subject to LiftGasBalance2{m in MANIFOLDSTOP, j in WELLSM[m]}:							# flow of liftgas to well j connected to topsideman m equals the flow routed to separators
	flow_LG[m,j] = sum{s in SEPARATORS}(topflow_LG[m,j,s]);
	
subject to GasFlowII{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]}:			# relationship between oil and gas flow in manifold routing(from well), subsea
	flow_manifold[m,j,l,'g'] = flow_manifold[m,j,l,'o']*GOR[m,j] + flow_LG_manifold[m,j,l];	

subject to WaterFlowII{m in MANIFOLDSSUB, j in WELLSM[m], l in PIPELINESM[m]}:			# relationship between oil and water flow in manifold routing, subsea
	flow_manifold[m,j,l,'w'] = flow_manifold[m,j,l,'o']*(WC[m,j]/(1-WC[m,j]));
	
subject to balanceCon5{m in MANIFOLDSTOP, j in WELLSM[m]}:								# flow out of well j equals the sum of flows to all separators from that well - extra constraint to make sure routing is feasible
	flow_well[m,j,'o'] = sum{s in SEPARATORS}topflow_sep[m,j,s,'o'];
	
subject to GasFlowIII{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS}:				# relationship between oil and gas flow in separator routing
	topflow_sep[m,j,s,'g'] = topflow_sep[m,j,s,'o']*GOR[m,j] + topflow_LG[m,j,s];	

subject to WaterFlowIII{m in MANIFOLDSTOP, j in WELLSM[m], s in SEPARATORS}:			# relationship between oil and water flow in separator routing
	topflow_sep[m,j,s,'w'] = topflow_sep[m,j,s,'o']*(WC[m,j]/(1-WC[m,j]));
	
subject to TotalGasCapacity:															#total gas flow through system must be less than then handling capacity
	sum{s in SEPARATORS, m in MANIFOLDSTOP, j in WELLSM[m]}topflow_sep[m,j,s,'g']
	+ sum{s in SEPARATORS, m in MANIFOLDSSUB, l in PIPELINESM[m]}subflow_sep[m,l,s,'g'] <= Gas_Capacity;
	
subject to SeparatorLiquidCapacity{s in SEPARATORS}:									#liquid flow must be less than the liquid handling capacity at each separator
	sum{m in MANIFOLDSTOP, j in WELLSM[m]}topflow_sep[m,j,s,'o']
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}subflow_sep[m,l,s,'o']
	+ sum{m in MANIFOLDSTOP, j in WELLSM[m]}topflow_sep[m,j,s,'w']
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}subflow_sep[m,l,s,'w'] <= Sep_Liquid_Capacity[s];
	
subject to SeparatorWaterCapacity{s in SEPARATORS}:										# water flow in to a separator  must be less than or equal to the separator
	sum{m in MANIFOLDSTOP, j in WELLSM[m]}topflow_sep[m,j,s,'w']
	+ sum{m in MANIFOLDSSUB, l in PIPELINESM[m]}subflow_sep[m,l,s,'w'] <= Sep_Water_Capacity[s];
	
